use serde::{Serialize, Deserialize};
use std::fs::File;
use std::io::Read;
use base64::prelude::*;

#[derive(Debug, PartialEq, Serialize, Deserialize, Clone)]
pub struct VPNConfig {
    public_key: String,
    private_key: String,
    endpoint: PeerEndpoint,
    ip: String,
    interface_name: String,
    namespace_name: String,
}

pub struct ConsumableVPNConfig {
    pub public_key: [u8; 32],
    pub private_key: [u8; 32],
    pub endpoint: ConsumablePeerEndpoint,
    pub ip: String,
    pub prefix: u8,
    pub interface_name: String,
    pub namespace_name: String,
}

#[derive(Debug, PartialEq, Serialize, Deserialize, Clone)]
pub struct PeerEndpoint {
    ip: String,
    port: u16,
}

pub struct ConsumablePeerEndpoint {
    pub ip: String,
    //pub prefix: u8,
    pub port: u16,
}

impl VPNConfig {
    pub fn get_consumable(self) -> ConsumableVPNConfig {
        let c_ip: Vec<&str> = self.ip.split("/").collect();
        if c_ip.len() != 2 {
            panic!("malformed ip, len:{}", c_ip.len());
        }
        let priv_key_dirty = BASE64_STANDARD.decode(self.private_key.into_bytes()).unwrap();
        let pub_key_dirty = BASE64_STANDARD.decode(self.public_key.into_bytes()).unwrap();
        let mut priv_key: [u8; 32] = Default::default();
        let mut pub_key: [u8; 32] = Default::default();
        priv_key.copy_from_slice(&priv_key_dirty[0..32]);
        pub_key.copy_from_slice(&pub_key_dirty[0..32]);
        ConsumableVPNConfig {
            public_key: pub_key,
            private_key: priv_key,
            endpoint: ConsumablePeerEndpoint {
                ip: self.endpoint.ip,
                port: self.endpoint.port,
            },
            ip: c_ip[0].to_string(),
            prefix: c_ip[1].parse::<u8>().unwrap(),
            interface_name: self.interface_name,
            namespace_name: self.namespace_name,
        }
    }
}

pub fn get_vpn_conf(file: String) -> Vec<VPNConfig> {
    let mut file = check_file(file);
    let mut s = String::new();
    file.read_to_string(&mut s).unwrap();
    match serde_yaml::from_str(&s) {
        Ok(result) => result,
        Err(e) => {
            log::error!("malformed: {}",e);
            Vec::new()
        },
    }
}

pub fn check_file(file: String) -> File {
    match File::open(file) {
       Ok(f) => f,
       Err(e) => {
        log::error!("Cannot file conf file: {}", e);
        panic!("{}", e);
        }
    }
}
