use nix::sched::{CloneFlags, unshare, setns};
use crate::namespace::consts::NET_NS_DIR;

use nix::fcntl::{open, OFlag};
use nix::mount::{mount, MsFlags};
use nix::unistd::{fork, ForkResult, Pid};
use nix::sys::wait::{waitpid, WaitStatus};
use nix::sys::stat::Mode;
use nix::sys::statvfs::{statvfs, FsFlags};

use std::path::{Path, PathBuf};
use std::process::exit;
use std::fs::File;
use std::os::fd::FromRawFd;

pub fn run_in_namespace<F>(f: F,ns_name: &String) -> Result<(), ()> where F:FnMut() + Copy {
    prep_for_fork()?;
    // Configure networking in the child namespace:
    // Fork a process that is set to the newly created namespace
    // Here set the veth ip addr, routing tables etc.
    // Unfortunately the NetworkNamespace interface of rtnetlink does
    // not offer these functionalities
    match unsafe { fork() } {
        Ok(ForkResult::Parent { child, .. }) => {
            // Parent process
            log::debug!("Net configuration PID: {}", child.as_raw());
            run_parent(child)
        }
        Ok(ForkResult::Child) => {
            // Child process
            // Move the child to the target namespace
            run_child(f, ns_name)
        }
        Err(e) => {
            log::error!("Can not fork() for ns creation: {}", e);
            return Err(());
        }
    }

}

fn run_parent(child: Pid) -> Result<(), ()> {
    log::trace!("[Parent] Child PID: {}", child);
    match waitpid(child, None) {
        Ok(wait_status) => match wait_status {
            WaitStatus::Exited(_, res) => {
                log::trace!("Child exited with: {}", res);
                if res == 0 {
                    return Ok(());
                } else {
                    log::error!("Child exited with status {}", res);
                    return Err(());
                }
            }
            WaitStatus::Signaled(_, _signal, _coredump) => {
                log::error!("Child process killed by signal");
                return Err(());
            }
            _ => {
                log::error!("Unknown child process status: {:?}", wait_status);
                return Err(());
            }
        }
        Err(e) => {
            log::error!("wait error : {}", e);
            return Err(());
        }
    }

}

fn run_child<F>(mut f: F, ns_name: &String) -> Result<(), ()> where F:FnMut() {
    let res = split_namespace(ns_name);

    match res {
        Err(_) => {
            log::error!("Child process crashed");
            std::process::abort()
        }
        Ok(()) => {
            log::debug!("Child exited normally");
            f();
            exit(0)
        }
    }
}

pub fn get_ns_fd(ns_name: &String) -> Result<File,()> {
    // Open NS path
    let ns_path = format!("{}/{}", NET_NS_DIR, ns_name);
    let mut open_flags = OFlag::empty();
    open_flags.insert(OFlag::O_RDONLY);
    open_flags.insert(OFlag::O_CLOEXEC);

    Ok(match open(Path::new(&ns_path), open_flags, Mode::empty()) {
        Ok(raw_fd) => unsafe {
            File::from_raw_fd(raw_fd)
        }
        Err(e) => {
            log::error!("Can not open network namespace: {}", e);
            return Err(());
        }
    })
}

fn split_namespace(ns_name: &String) -> Result<(), ()> {
    let fd = get_ns_fd(ns_name)?;
    // Switch to network namespace with CLONE_NEWNET
    if let Err(e) = setns(fd, CloneFlags::CLONE_NEWNET) {
        log::error!("Can not set namespace to target {}: {}", ns_name, e);
        return Err(());
    }
    // unshare with CLONE_NEWNS
    if let Err(e) = unshare(CloneFlags::CLONE_NEWNS) {
        log::error!("Can not unshare: {}", e);
        return Err(());
    }
    // mount blind the fs
    // let's avoid that any mount propagates to the parent process
    // mount_directory(None, &PathBuf::from("/"), vec![MsFlags::MS_REC, MsFlags::MS_PRIVATE])?;
    let mut mount_flags = MsFlags::empty();
    mount_flags.insert(MsFlags::MS_REC);
    mount_flags.insert(MsFlags::MS_PRIVATE);
    if let Err(_e) = mount::<PathBuf, PathBuf, str, PathBuf>(None, &PathBuf::from("/"), None, mount_flags, None) {
        log::error!("Can not remount root directory");
        ()
    }

    // Now unmount /sys
    let sys_path = PathBuf::from("/sys");
    mount_flags = MsFlags::empty();
    // Needed to respect the trait for NixPath
    let ns_name_path = PathBuf::from(ns_name);

    // TODO do not exit for EINVAL error
    // unmount_path(&sys_path)?;
    // consider the case that a sysfs is not present
    let stat_sys = statvfs(&sys_path)
        .map_err(|e| {
            log::error!("Can not stat sys: {}", e);
    }).unwrap();
    if stat_sys.flags().contains(FsFlags::ST_RDONLY) {
        mount_flags.insert(MsFlags::MS_RDONLY);
    }

    // and remount a version of /sys that describes the network namespace
    if let Err(e) = mount::<PathBuf, PathBuf, str, PathBuf>(Some(&ns_name_path), &sys_path, Some("sysfs"), mount_flags, None) {
        log::error!("Can not remount /sys to namespace: {}", e);
        ()
    }

    Ok(())
}


// Cargo cult from the definition in rtnetlink
fn prep_for_fork() -> Result<(), ()> {
    Ok(())
}
